# audio_controller.py
import sounddevice as sd
import numpy as np
from typing import Optional

class AudioController:
    def __init__(self, audio: Optional[np.ndarray] = None, samplerate: Optional[float] = None):
        self.audio = audio
        self.samplerate = samplerate
        self.stream = None
        self.current_frame = 0
        self.blocksize = 1024
        self.playing = False
        self.paused = False

    def set_params(self, audio: np.ndarray, samplerate: float):
        self.audio = audio
        self.samplerate = samplerate
        self.current_frame = 0
        self.playing = False
        self.paused = False
        if self.stream:
            self.stream.close()
            self.stream = None

    def callback(self, outdata: np.ndarray, frames: int, time, status):
        if not self.playing or self.audio is None:
            outdata[:] = np.zeros((frames, 1))
            raise sd.CallbackStop

        chunk = self.audio[self.current_frame:self.current_frame + frames]
        if len(chunk) < frames:
            chunk = np.pad(chunk, (0, frames - len(chunk)))

        outdata[:] = chunk.reshape(-1, 1)
        self.current_frame += frames

        #if self.current_frame >= len(self.audio):
        #    self.playing = False
        #    raise sd.CallbackStop

    def reset(self):
        self.playing = False
        self.paused = False
        self.current_frame = 0
        self.stream.stop()
        #if self.stream:
        #    self.stream.stop()
        #    self.stream.close()
        #    self.stream = None


    def pause(self):
        if self.stream and self.playing:
            self.stream.stop()
            self.paused = True
            self.playing = False

    def resume(self):
        if self.audio is None or self.samplerate is None:
            raise ValueError("Audio and Sample Rate must be set before playback")

        if not self.stream:
            self.current_frame = 0
            self.stream = sd.OutputStream(samplerate=self.samplerate,
                                          channels=1,
                                          callback=self.callback,
                                          blocksize=self.blocksize)

        if not self.playing:
            self.playing = True
            self.paused = False
            self.stream.start()



import numpy as np
from typing import Optional, Tuple
import cv2
from visound.core.TraversalMode import TraversalMode

class Sonify:
    def __init__(self,
                 file_path: str,
                 dimension: Tuple[int, int] = (128, 128),
                 duration_per_column: Optional[float] = 0.01,
                 sample_rate: Optional[float] = 44100):

        self.file_path = file_path
        self.dim = dimension
        self.DPC = duration_per_column
        self.SR = sample_rate
        self.height = self.dim[0]
        self.width = self.dim[1]
        self.traversal_mode = None

        self.image = cv2.imread(self.file_path, cv2.IMREAD_GRAYSCALE)

        if self.image is None:
            raise FileNotFoundError(f"Imafe file not found or unreadable: {self.file_path}")
        self.image = cv2.resize(self.image, self.dim)



        # from core.AudioController import AudioController
        # self.audio_controller = AudioController()

    # def GUI(self) -> None:
    #     import sys
    #     from PyQt6.QtWidgets import QApplication
    #     from core.GUI import MainWindow

    #     app = QApplication(sys.argv)
    #     GUI = MainWindow()
    #     GUI.set_traversal_mode(self.traversal_mode)
    #     GUI.loadImage(self.image)
    #     GUI.dpc = self.DPC
    #     GUI.init_bar_position()
    #     GUI.reset_signal.connect(self.reset)
    #     GUI.pause_resume_signal.connect(self.pause_or_resume)
    #     app.exec()

    # def pause_or_resume(self, pause: bool) -> None:
    #     if pause:
    #         self.audio_controller.resume()
    #     else:
    #         self.audio_controller.pause()

    # def reset(self) -> None:
    #     self.audio_controller.reset()

    def pixel_to_freq(self, y: float, height: float) -> float:
        """
        Mapping function of pixel to frequency
        """
        return 200 + (1 - y / height) * 1800

    def LTR(self) -> np.ndarray:
        """
        Left to Right traversal of image
        """
        if self.image is None:
            raise ValueError("No image loaded to sonify.")

        self.traversal_mode = TraversalMode.LeftToRight

        sound = np.zeros(int(self.width * self.DPC * self.SR))
        t_col = np.linspace(0, self.DPC, int(self.DPC * self.SR), endpoint=False)

        for x in range(self.width):
            column = self.image[:, x]
            column_sound = np.zeros_like(t_col)

            for y in range(self.height):
                intensity = column[y] / 255.0
                if intensity > 0.1:
                    freq = self.pixel_to_freq(y, self.height)
                    column_sound += intensity * np.sin(2 * np.pi * freq * t_col)

            start = int(x * self.DPC * self.SR)
            end = start + len(t_col)
            sound[start:end] += column_sound

       # self.audio_controller.set_params(sound, self.SR)

        return sound

    def RTL(self) -> np.ndarray:
        """
        Right to Left traversal of image
        """

        if self.image is None:
            raise ValueError("No image loaded to sonify.")

        self.traversal_mode = TraversalMode.RightToLeft

        sound = np.zeros(int(self.width * self.DPC * self.SR))
        t_col = np.linspace(0, self.DPC, int(self.DPC * self.SR), endpoint=False)

        for i, x in enumerate(range(self.width - 1, -1, -1)):
            column = self.image[:, x]
            column_sound = np.zeros_like(t_col)

            for y in range(self.height):
                intensity = column[y] / 255.0
                if intensity > 0.1:
                    freq = self.pixel_to_freq(y, self.height)
                    column_sound += intensity * np.sin(2 * np.pi * freq * t_col)

            start = int(i * self.DPC * self.SR)
            end = start + len(t_col)
            sound[start:end] += column_sound

        # self.audio_controller.set_params(sound, self.SR)

        return sound
